from AOT_biomaps.AOT_Recon.ReconTools import power_method, gradient_cpu, gradient_gpu, div_cpu, div_gpu, proj_l2, gradient_KL
from AOT_biomaps.Config import config

import torch
from tqdm import trange

def chambolle_pock_TV_cpu(SMatrix, y, alpha, theta, numIterations, isSavingEachIteration, L, withTumor):
    device = torch.device("cpu")
    A = torch.tensor(SMatrix, dtype=torch.float32, device=device)
    y = torch.tensor(y, dtype=torch.float32, device=device)

    T, Z, X, N = SMatrix.shape
    A_flat = A.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y.reshape(-1)

    P = lambda x: torch.matmul(A_flat, x.ravel())
    PT = lambda y: torch.matmul(A_flat.T, y)

    if L is None:
        L = power_method(P, PT, y_flat, Z, X, isGPU=True)

    sigma = 1.0 / L
    tau = 1.0 / L

    x = torch.zeros(Z * X, device=device)
    p = torch.zeros((2, Z, X), device=device)
    q = torch.zeros_like(y_flat)
    x_tilde = x.clone()

    I_reconMatrix = [x.reshape(Z, X).cpu().numpy()]

    if withTumor:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (TV : Gaussian Noise) α:{alpha:.4f} L: {L:.4f} ---- WITH TUMOR ---- processing on single CPU ----"
    else:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (TV : Gaussian Noise) α:{alpha:.4f} L: {L:.4f} ---- WITHOUT TUMOR ---- processing on single CPU ----"

    for iteration in trange(numIterations, desc=description):
        p = proj_l2(p + sigma * gradient_cpu(x_tilde.reshape(Z, X)), alpha)
        q = (q + sigma * P(x_tilde) - sigma * y_flat) / (1.0 + sigma)

        x_old = x
        x = x + tau * div_cpu(p).ravel() - tau * PT(q)
        x_tilde = x + theta * (x - x_old)

        if iteration % 1 == 0:
            I_reconMatrix.append(x.reshape(Z, X).cpu().numpy())

    return I_reconMatrix if isSavingEachIteration else I_reconMatrix[-1]


def chambolle_pock_TV_gpu(SMatrix, y, alpha, theta, numIterations, isSavingEachIteration, L, withTumor):

    device = torch.device(f"cuda:{config.select_best_gpu()}")
    A = torch.tensor(SMatrix, dtype=torch.float32, device=device)
    y = torch.tensor(y, dtype=torch.float32, device=device)
    T, Z, X, N = SMatrix.shape
    A_flat = A.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y.reshape(-1)
    P = lambda x: torch.matmul(A_flat, x.ravel())
    PT = lambda y: torch.matmul(A_flat.T, y)

    if L is None:
        L = power_method(P, PT, y_flat, Z, X, isGPU=True)

    sigma = 1.0 / L
    tau = 1.0 / L
    x = torch.zeros(Z * X, device=device)
    p = torch.zeros((2, Z, X), device=device)
    q = torch.zeros_like(y_flat)
    x_tilde = x.clone()
    I_reconMatrix = [x.reshape(Z, X).cpu().numpy()]

    if withTumor:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (TV : Gaussian Noise) α:{alpha:.4f} L: {L:.4f} ---- WITH TUMOR ---- processing on GPU no.{torch.cuda.current_device()} ----"
    else:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (TV: Gaussian Noise) α:{alpha:.4f} L: {L:.4f} ---- WITHOUT TUMOR ---- processing on GPU no.{torch.cuda.current_device()} ----"

    for iteration in trange(numIterations, desc=description):
        p = proj_l2(p + sigma * gradient_gpu(x_tilde.reshape(Z, X)), alpha)
        q = (q + sigma * P(x_tilde) - sigma * y_flat) / (1.0 + sigma)
        x_old = x
        x = x + tau * div_gpu(p).ravel() - tau * PT(q)
        x_tilde = x + theta * (x - x_old)
        if iteration % 1 == 0:
            I_reconMatrix.append(x.reshape(Z, X).cpu().numpy())
    return I_reconMatrix if isSavingEachIteration else I_reconMatrix[-1]

def chambolle_pock_KL_cpu(SMatrix, y, alpha, theta, numIterations, isSavingEachIteration, L, withTumor):

    device = torch.device("cpu")
    A = torch.tensor(SMatrix, dtype=torch.float32, device=device)
    y = torch.tensor(y, dtype=torch.float32, device=device)
    T, Z, X, N = SMatrix.shape
    A_flat = A.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y.reshape(-1)

    P = lambda x: torch.matmul(A_flat, x.ravel())
    PT = lambda y: torch.matmul(A_flat.T, y)

    # Initialisation
    x = torch.ones(Z * X, device=device)  # Initialisation positive pour KL
    p = torch.zeros((2, Z, X), device=device)
    q = torch.zeros_like(y_flat)
    x_tilde = x.clone()

    # Paramètres de l'algorithme
    sigma = 0.1
    tau = 0.1

    I_reconMatrix = [x.reshape(Z, X).cpu().numpy()]

    if L is None:
        L = power_method(P, PT, y_flat, Z, X, isGPU=True)

    if withTumor:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (KL : Poisson noise) α:{alpha:.4f} L: {L:.4f} ---- WITH TUMOR ---- processing on single CPU ----"
    else:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (KL : Poisson noise) α:{alpha:.4f} L: {L:.4f} ---- WITHOUT TUMOR ---- processing on single CPU ----"
    
    for iteration in trange(numIterations, desc= description):
        # Mise à jour de p (gradient de la variation totale)
        p = proj_l2(p + sigma * gradient_cpu(x_tilde.reshape(Z, X)), alpha)

        # Mise à jour de q (gradient de la divergence de Kullback-Leibler)
        Ax = P(x_tilde)
        q = q + sigma * gradient_KL(Ax, y_flat)

        # Mise à jour de x
        x_old = x
        x = x - tau * (div_cpu(p).ravel() + PT(q))
        x = torch.clamp(x, min=1e-10)  # Assurez-vous que x reste positif
        x_tilde = x + theta * (x - x_old)

        if iteration % 1 == 0:
            I_reconMatrix.append(x.reshape(Z, X).cpu().numpy())

    return I_reconMatrix if isSavingEachIteration else I_reconMatrix[-1]

def chambolle_pock_KL_gpu(SMatrix, y, alpha, theta, numIterations, isSavingEachIteration, L, withTumor):
    device = torch.device(f"cuda:{config.select_best_gpu()}")
    A = torch.tensor(SMatrix, dtype=torch.float32, device=device)
    y = torch.tensor(y, dtype=torch.float32, device=device)
    T, Z, X, N = SMatrix.shape
    A_flat = A.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y.reshape(-1)

    P = lambda x: torch.matmul(A_flat, x.ravel())
    PT = lambda y: torch.matmul(A_flat.T, y)

    if L is None:
        L = power_method(P, PT, y_flat, Z, X, isGPU=True)

    # Initialisation
    x = torch.ones(Z * X, device=device)  # Initialisation positive pour KL
    p = torch.zeros((2, Z, X), device=device)
    q = torch.zeros_like(y_flat)
    x_tilde = x.clone()


    # Paramètres de l'algorithme
    sigma = 0.1
    tau = 0.1

    I_reconMatrix = [x.reshape(Z, X).cpu().numpy()]
    
    if withTumor:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (KL : Poisson noise) α:{alpha:.4f} L: {L:.4f} ---- WITH TUMOR  ---- processing on GPU no.{torch.cuda.current_device()} ----"
    else:
        description = f"AOT-BioMaps -- Primal/Dual Recontruction Tomography : Chambolle-Pock (KL : Poisson noise) α:{alpha:.4f} L: {L:.4f} ---- WITHOUT TUMOR  ---- processing on GPU no.{torch.cuda.current_device()} ----"

    for iteration in trange(numIterations, desc=description):
        # Mise à jour de p (gradient de la variation totale)
        p = proj_l2(p + sigma * gradient_gpu(x_tilde.reshape(Z, X)), alpha)

        # Mise à jour de q (gradient de la divergence de Kullback-Leibler)
        Ax = P(x_tilde)
        q = q + sigma * gradient_KL(Ax, y_flat)

        # Mise à jour de x
        x_old = x
        x = x - tau * (div_gpu(p).ravel() + PT(q))
        x = torch.clamp(x, min=1e-10)  # Assurez-vous que x reste positif
        x_tilde = x + theta * (x - x_old)

        if iteration % 1 == 0:
            I_reconMatrix.append(x.reshape(Z, X).cpu().numpy())

    return I_reconMatrix if isSavingEachIteration else I_reconMatrix[-1]



from AOT_biomaps.Config import config

import torch
import numpy as np
from numba import njit, prange
import torch.nn.functional as F
if config.get_process()  == 'gpu':
    try:
        from torch_sparse import coalesce
    except ImportError:
        raise ImportError("torch_scatter and torch_sparse are required for GPU processing. Please install them using 'pip install torch-scatter torch-sparse' with correct link (follow instructions https://github.com/LucasDuclos/AcoustoOpticTomography/edit/main/README.md).")


@njit(parallel=True)
def _forward_projection(SMatrix, theta_p, q_p):
    t_dim, z_dim, x_dim, i_dim = SMatrix.shape
    for _t in prange(t_dim):
        for _n in range(i_dim):
            total = 0.0
            for _z in range(z_dim):
                for _x in range(x_dim):
                    total += SMatrix[_t, _z, _x, _n] * theta_p[_z, _x]
            q_p[_t, _n] = total

@njit(parallel=True)
def _backward_projection(SMatrix, e_p, c_p):
    t_dim, z_dim, x_dim, n_dim = SMatrix.shape
    for _z in prange(z_dim):
        for _x in range(x_dim):
            total = 0.0
            for _t in range(t_dim):
                for _n in range(n_dim):
                    total += SMatrix[_t, _z, _x, _n] * e_p[_t, _n]
            c_p[_z, _x] = total

@njit
def _build_adjacency_sparse_CPU(Z, X,corner = (0.5-np.sqrt(2)/4)/np.sqrt(2),face = 0.5-np.sqrt(2)/4):
    rows = []
    cols = []
    weights = []

    for z in range(Z):
        for x in range(X):
            j = z * X + x
            for dz in [-1, 0, 1]:
                for dx in [-1, 0, 1]:
                    if dz == 0 and dx == 0:
                        continue
                    nz, nx = z + dz, x + dx
                    if 0 <= nz < Z and 0 <= nx < X:
                        k = nz * X + nx
                        weight = corner if abs(dz) + abs(dx) == 2 else face
                        rows.append(j)
                        cols.append(k)
                        weights.append(weight)

    index = (np.array(rows), np.array(cols))
    values = np.array(weights, dtype=np.float32)
    return index, values 

def _build_adjacency_sparse_GPU(Z, X,corner = (0.5-np.sqrt(2)/4)/np.sqrt(2),face = 0.5-np.sqrt(2)/4):
    rows = []
    cols = []
    weights = []

    for z in range(Z):
        for x in range(X):
            j = z * X + x
            for dz in [-1, 0, 1]:
                for dx in [-1, 0, 1]:
                    if dz == 0 and dx == 0:
                        continue
                    nz, nx = z + dz, x + dx
                    if 0 <= nz < Z and 0 <= nx < X:
                        k = nz * X + nx
                        weight = corner if abs(dz) + abs(dx) == 2 else face
                        rows.append(j)
                        cols.append(k)
                        weights.append(weight)

    index = torch.tensor([rows, cols], device= config.select_best_gpu())
    values = torch.tensor(weights, dtype=torch.float32, device= config.select_best_gpu())
    index, values = coalesce(index, values, m=Z*X, n=Z*X)
    return index, values 


def power_method(P, PT, data, Z, X, n_it=10, isGPU=False):
    x = PT(data)
    x = x.reshape(Z, X)
    for _ in range(n_it):
        grad = gradient_gpu(x) if isGPU else gradient_cpu(x)
        div = div_gpu(grad) if isGPU else div_cpu(grad)
        x = PT(P(x.ravel())) - div.ravel()
        s = torch.sqrt(torch.sum(x**2))
        x /= s
        x = x.reshape(Z, X)
    return torch.sqrt(s)

def proj_l2(p, alpha):
    norm = torch.sqrt(torch.sum(p**2, dim=0, keepdim=True))
    return p * alpha / torch.max(norm, torch.tensor(alpha, device=p.device))

def norm2sq(x):
    return torch.sum(x**2)

def norm1(x):
    return torch.sum(torch.abs(x))

def gradient_cpu(x):
    grad_x = torch.zeros_like(x)
    grad_y = torch.zeros_like(x)

    grad_x[:-1, :] = x[1:, :] - x[:-1, :]
    grad_y[:, :-1] = x[:, 1:] - x[:, :-1]

    return torch.stack((grad_x, grad_y), dim=0)

def div_cpu(x):
    if x.dim() == 3:
        x = x.unsqueeze(0)  # Devient [1, 2, H, W]

    gx = x[:, 0:1, :, :]  # gradient horizontal
    gy = x[:, 1:2, :, :]  # gradient vertical

    # Définition des noyaux de divergence
    kernel_x = torch.tensor([[[[1.0], [-1.0]]]], dtype=torch.float32)
    kernel_y = torch.tensor([[[[1.0, -1.0]]]], dtype=torch.float32)

    # Appliquer la convolution
    div_x = F.conv2d(gx, kernel_x, padding=(1, 0))
    div_y = F.conv2d(gy, kernel_y, padding=(0, 1))

    # Rogner pour avoir la même taille (H, W)
    H, W = x.shape[2:]
    div_x = div_x[:, :, :H, :]
    div_y = div_y[:, :, :, :W]

    return -(div_x + div_y).squeeze()

def gradient_gpu(x):
    grad_x = torch.zeros_like(x)
    grad_y = torch.zeros_like(x)
    grad_x[:-1, :] = x[1:, :] - x[:-1, :]
    grad_y[:, :-1] = x[:, 1:] - x[:, :-1]
    return torch.stack((grad_x, grad_y), dim=0)

def div_gpu(x):
    if x.dim() == 3:
        x = x.unsqueeze(0)  # Devient [1, 2, H, W]
    gx = x[:, 0:1, :, :]  # gradient horizontal
    gy = x[:, 1:2, :, :]  # gradient vertical

    # Définition des noyaux de divergence
    kernel_x = torch.tensor([[[[1.0], [-1.0]]]], dtype=torch.float32, device=x.device)
    kernel_y = torch.tensor([[[[1.0, -1.0]]]], dtype=torch.float32, device=x.device)

    # Appliquer la convolution
    div_x = F.conv2d(gx, kernel_x, padding=(1, 0))
    div_y = F.conv2d(gy, kernel_y, padding=(0, 1))

    # Rogner pour avoir la même taille (H, W)
    H, W = x.shape[2:]
    div_x = div_x[:, :, :H, :]
    div_y = div_y[:, :, :, :W]

    return -(div_x + div_y).squeeze()

def KL_divergence(Ax, y):
    return torch.sum(Ax - y * torch.log(Ax + 1e-10))

def gradient_KL(Ax, y):
    return 1 - y / (Ax + 1e-10)


# ACOUSTIC
from .AOT_Acoustic._mainAcoustic import *
from .AOT_Acoustic.AcousticEnums import *
from .AOT_Acoustic.AcousticTools import *
from .AOT_Acoustic.FocusedWave import *
from .AOT_Acoustic.IrregularWave import *
from .AOT_Acoustic.PlaneWave import *
from .AOT_Acoustic.StructuredWave import *

# EXPERIMENT
from .AOT_Experiment._mainExperiment import *
from .AOT_Experiment.Focus import *
from .AOT_Experiment.Tomography import *

# OPTIC
from .AOT_Optic._mainOptic import *
from .AOT_Optic.Absorber import *
from .AOT_Optic.Laser import *
from .AOT_Optic.OpticEnums import *

# RECONSTRUCTION
from .AOT_Recon._mainRecon import *
from .AOT_Recon.AlgebraicRecon import *
from .AOT_Recon.AnalyticRecon import *
from .AOT_Recon.BayesianRecon import *
from .AOT_Recon.DeepLearningRecon import *
from .AOT_Recon.PrimalDualRecon import *
from .AOT_Recon.ReconEnums import *
from .AOT_Recon.ReconTools import *

# OPTIMIZERS
from .AOT_Recon.AOT_Optimizers.DEPIERRO import *
from .AOT_Recon.AOT_Optimizers.MAPEM import *
from .AOT_Recon.AOT_Optimizers.MLEM import *
from .AOT_Recon.AOT_Optimizers.PDHG import *

# POTENTIAL FUNCTIONS
from .AOT_Recon.AOT_PotentialFunctions.Huber import *
from .AOT_Recon.AOT_PotentialFunctions.Quadratic import *
from .AOT_Recon.AOT_PotentialFunctions.RelativeDifferences import *

# CONFIG AND SETTINGS
from .Config import config
from .Settings import *

__version__ = '2.7.9'

if config.get_process() == 'gpu':
    __process__ = 'gpu'
else:
    __process__ = 'cpu'

def initialize(process='cpu'):
    config.set_process(process)
    if process == 'gpu':
        config.select_best_gpu()
        print(f"Initialized with process: {config.get_process()} using GPU: {config.bestGPU}")
    else:
        print(f"Initialized with process: {config.get_process()}")



from ._mainExperiment import Experiment
from AOT_biomaps.AOT_Acoustic.AcousticEnums import WaveType
from AOT_biomaps.AOT_Acoustic.StructuredWave import StructuredWave
from AOT_biomaps.Config import config

import os
import psutil
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import trange

class Tomography(Experiment):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    # PUBLIC METHODS
        
    def check(self):
        """
        Check if the experiment is correctly initialized.
        """
        if self.TypeAcoustic is None or self.TypeAcoustic.value == WaveType.FocusedWave.value:
           return False, "acousticType must be provided and cannot be FocusedWave for Tomography experiment"
        if self.AcousticFields is None:
           return False, "AcousticFields is not initialized. Please generate the system matrix first."
        if self.AOsignal_withTumor is None:
            return False, "AOsignal with tumor is not initialized. Please generate the AO signal with tumor first."   
        if self.AOsignal_withoutTumor is None:
            return False, "AOsignal without tumor is not initialized. Please generate the AO signal without tumor first." 
        if self.OpticImage is None:
            return False, "OpticImage is not initialized. Please generate the optic image first."
        if self.AOsignal_withoutTumor.shape != self.AOsignal_withTumor.shape:
            return False, "AOsignal with and without tumor must have the same shape."
        for field in self.AcousticFields:
            if field.field.shape[0] != self.AOsignal_withTumor.shape[0]:
                return False, f"Field {field.getName_field()} has an invalid Time shape: {field.field.shape[0]}. Expected time shape to be {self.AOsignal_withTumor.shape[0]}."
        if not all(field.field.shape == self.AcousticFields[0].field.shape for field in self.AcousticFields):
            return False, "All AcousticFields must have the same shape."
        if self.OpticImage is None:
            return False, "OpticImage is not initialized. Please generate the optic image first."
        if self.OpticImage.phantom is None:
            return False, "OpticImage phantom is not initialized. Please generate the phantom first."
        if self.OpticImage.laser is None:
            return False, "OpticImage laser is not initialized. Please generate the laser first."
        if self.OpticImage.laser.shape != self.OpticImage.phantom.shape:
            return False, "OpticImage laser and phantom must have the same shape."
        if self.OpticImage.phantom.shape[0] != self.AcousticFields[0].field.shape[1] or self.OpticImage.phantom.shape[1] != self.AcousticFields[0].field.shape[2]:
            return False, f"OpticImage phantom shape {self.OpticImage.phantom.shape} does not match AcousticFields shape {self.AcousticFields[0].field.shape[1:]}."
        
        return True, "Experiment is correctly initialized."

    def generateAcousticFields(self, fieldDataPath, fieldParamPath, show_log = True):
        """
        Generate the acoustic fields for simulation.

        Args:
            fieldDataPath: Path to save the generated fields.
            fieldParamPath: Path to the field parameters file.

        Returns:
            systemMatrix: A numpy array of the generated fields.
        """
        if self.TypeAcoustic.value == WaveType.StructuredWave.value:
            self.AcousticFields = self._generateAcousticFields_STRUCT_CPU(fieldDataPath, fieldParamPath,show_log)
        else:
            raise ValueError("Unsupported wave type.")

    def show_pattern(self):
        if self.AcousticFields is None:
            raise ValueError("AcousticFields is not initialized. Please generate the system matrix first.")

        entries = []
        for field in self.AcousticFields:
            if field.waveType != WaveType.StructuredWave:
                raise TypeError("AcousticFields must be of type StructuredWave to plot pattern.")
            entries.append(((field.pattern.space_0, field.pattern.space_1, field.pattern.move_head_0_2tail, field.pattern.move_tail_1_2head), field.pattern.activeList, field.angle))
        print("Entries:", entries)
        

        # Sorting rule
        entries.sort(
            key=lambda x: (
                -(x[0][0] + x[0][1]),  # Total length descending
                -max(x[0][0], x[0][1]), # Max(space_0, space_1) descending
                -x[0][0],              # space_0 descending
                -x[0][2],              # move_head_0_2tail descending
                x[0][3]                # move_tail_1_2head ascending
            )
        )

        df = pd.DataFrame([
            {
                "hex": hex_str,
                "space_0": t[0],
                "space_1": t[1],
                "move_head_0_2tail": t[2],
                "move_tail_1_2head": t[3],
                "angles": angles
            }
            for t, hex_str, angles in entries
        ])

        def hex_string_to_binary_column(hex_str):
            bits = ''.join(f'{int(c, 16):04b}' for c in hex_str)
            return np.array([int(b) for b in bits], dtype=np.uint8).reshape(-1, 1)

        hex_list = df['hex'].tolist()
        angle_list = df['angles'].tolist()
        bit_columns = [hex_string_to_binary_column(h) for h in hex_list]
        image = np.hstack(bit_columns)
        height = image.shape[0]

        _, ax = plt.subplots(figsize=(12, 10))
        ax.imshow(image, cmap='gray', aspect='auto')
        ax.set_title("Scan configuration", fontsize='large')
        ax.set_xlabel("Wave", fontsize='medium')
        ax.set_ylabel("Transducer activation", fontsize='medium')

        angle_min = -20.2
        angle_max = 20.2
        center = height / 2
        scale = height / (angle_max - angle_min)

        for i, angle in enumerate(angle_list):
            y = round(center - angle * scale)
            if 0 <= y <= height:
                ax.plot(i, y-0.5, 'r.', markersize=5)

        ax.set_ylim(height - 0.5, -0.5)

        ax2 = ax.twinx()
        ax2.set_ylim(ax.get_ylim())

        yticks_angle = np.linspace(20, -20, 9)
        yticks_pos = np.interp(yticks_angle, [angle_min, angle_max], [height - 0.5, -0.5])
        ax2.set_yticks(yticks_pos)
        ax2.set_yticklabels([f"{a:.1f}°" for a in yticks_angle])
        ax2.set_ylabel("Angle [degree]", fontsize='medium', color='r')
        ax2.tick_params(axis='y', colors='r')

        plt.show()
     
    # PRIVATE METHODS

    def _generateAcousticFields_STRUCT_CPU(self, fieldDataPath, fieldParamPath, show_log):
        if not os.path.exists(fieldParamPath):
            raise FileNotFoundError(f"Field parameter file {fieldParamPath} not found.")
        if not fieldDataPath is None:
            os.makedirs(fieldDataPath, exist_ok=True)
        listAcousticFields = []
        patternList = []
        with open(fieldParamPath, 'r') as file:
            lines = file.readlines()
            for line in lines:
                line = line.strip()
                if not line:
                    continue  # skip empty lines

                try:
                    # Sécurise l'évaluation en supprimant accès à builtins
                    parsed = eval(line, {"__builtins__": None})

                    if isinstance(parsed, tuple) and len(parsed) == 2:
                        coords, angles = parsed
                        for angle in angles:
                            patternList.append([*coords, angle])
                    else:
                        raise ValueError("Ligne inattendue (pas un tuple de deux éléments)")

                except Exception as e:
                    print(f"Erreur de parsing sur la ligne : {line}\n{e}")

        progress_bar = trange(0,len(patternList), desc="Generating acoustic fields")

        for i in progress_bar:
            memory = psutil.virtual_memory()
            pattern = patternList[i]
            if len(pattern) != 5:
                raise ValueError(f"Invalid pattern format: {pattern}. Expected 5 values.")
            # Initialisation de l'objet AcousticField
            AcousticField = StructuredWave(
                angle_deg=pattern[4],
                space_0=pattern[0],
                space_1=pattern[1],
                move_head_0_2tail=pattern[2],
                move_tail_1_2head=pattern[3],
                params=self.params
            )

            if fieldDataPath is None:
                pathField = None
            else:
                pathField = os.path.join(fieldDataPath, os.path.basename(AcousticField.getName_field() + self.FormatSave.value))

            if not pathField is None and os.path.exists(pathField):
                if progress_bar is not None:
                    progress_bar.set_postfix_str(f"Loading field - {AcousticField.getName_field()} -- Memory used :{memory.percent}%")
                    try:
                        AcousticField.load_field(fieldDataPath,  self.FormatSave)
                    except:
                        progress_bar.set_postfix_str(f"Error loading field -> Generating field - {AcousticField.getName_field()} -- Memory used :{memory.percent}% ---- processing on {config.get_process().upper()} ----")
                        AcousticField.generate_field(show_log = show_log)
                        if not pathField is None and not os.path.exists(pathField):
                            progress_bar.set_postfix_str(f"Saving field - {AcousticField.getName_field()} -- Memory used :{memory.percent}%")
                            os.makedirs(os.path.dirname(pathField), exist_ok=True) 
                            AcousticField.save_field(fieldDataPath)

            elif pathField is None or not os.path.exists(pathField):
                progress_bar.set_postfix_str(f"Generating field - {AcousticField.getName_field()} -- Memory used :{memory.percent}% ---- processing on {config.get_process().upper()} ----")
                AcousticField.generate_field(show_log = show_log)
            
            if not pathField is None and not os.path.exists(pathField):
                progress_bar.set_postfix_str(f"Saving field - {AcousticField.getName_field()} -- Memory used :{memory.percent}%")
                os.makedirs(os.path.dirname(pathField), exist_ok=True) 
                AcousticField.save_field(fieldDataPath)

            listAcousticFields.append(AcousticField)
            # Réinitialiser le texte de la barre de progression pour l'itération suivante
            progress_bar.set_postfix_str("")
   
        return listAcousticFields
    



from AOT_biomaps.Config import config
from AOT_biomaps.AOT_Experiment.Tomography import Tomography

import os
import numpy as np
from abc import ABC, abstractmethod
from .ReconEnums import ReconType
from sklearn.metrics import mean_squared_error
from skimage.metrics import structural_similarity as ssim

class Recon(ABC):
    def __init__(self, experiment, saveDir, isGPU = config.get_process() == 'gpu',  isMultiGPU =  True if config.numGPUs > 1 else False, isMultiCPU = True):
        self.reconPhantom = None
        self.reconLaser = None
        self.experiment = experiment
        self.reconType = None
        self.saveDir = saveDir
        self.MSE = None
        self.SSIM = None

        self.isGPU = isGPU
        self.isMultiGPU = isMultiGPU
        self.isMultiCPU = isMultiCPU

        if str(type(self.experiment)) != str(Tomography):
            raise TypeError(f"Experiment must be of type {Tomography}")

    @abstractmethod
    def run(self,withTumor = True):
        pass

    def calculateCRC(self,iteration,ROI_mask = None):
        """
        Computes the Contrast Recovery Coefficient (CRC) for a given ROI.
        """
        if self.reconType is ReconType.Analytic:
            raise TypeError(f"Impossible to calculate CRC with analytical reconstruction")
        elif self.reconType is None:
            raise ValueError("Run reconstruction first")
        
        if self.reconLaser is None or self.reconLaser == []:
            raise ValueError("Reconstructed laser is empty. Run reconstruction first.")
        if isinstance(self.Laser,list) and len(self.Laser) == 1:
            raise ValueError("Reconstructed Image without tumor is a single frame. Run reconstruction with isSavingEachIteration=True to get a sequence of frames.")
        if self.reconPhantom is None or self.reconPhantom == []:
            raise ValueError("Reconstructed phantom is empty. Run reconstruction first.")
        if isinstance(self.reconPhantom, list) and len(self.reconPhantom) == 1:
            raise ValueError("Reconstructed Image with tumor is a single frame. Run reconstruction with isSavingEachIteration=True to get a sequence of frames.")
        
        if self.reconLaser is None or self.reconLaser == []:
            print("Reconstructed laser is empty. Running reconstruction without tumor...")
            self.run(withTumor = False, isSavingEachIteration=True)
        if ROI_mask is not None:
            recon_ratio = np.mean(self.reconPhantom[iteration][ROI_mask]) / np.mean(self.reconLaser[iteration][ROI_mask])
            lambda_ratio = np.mean(self.experiment.OpticImage.phantom[ROI_mask]) / np.mean(self.experiment.OpticImage.laser[ROI_mask]) 
        else:
            recon_ratio = np.mean(self.reconPhantom[iteration]) / np.mean(self.reconLaser[iteration])
            lambda_ratio = np.mean(self.experiment.OpticImage.phantom) / np.mean(self.experiment.OpticImage.laser)
        
        # Compute CRC
        CRC = (recon_ratio - 1) / (lambda_ratio - 1)
        return CRC
    
    def calculateMSE(self):
        """
        Calculate the Mean Squared Error (MSE) of the reconstruction.

        Returns:
            mse: float or list of floats, Mean Squared Error of the reconstruction
        """
                
        if self.reconPhantom is None or self.reconPhantom == []:
            raise ValueError("Reconstructed phantom is empty. Run reconstruction first.")

        if self.reconType in (ReconType.Analytic, ReconType.DeepLearning):
            self.MSE = mean_squared_error(self.experiment.OpticImage.phantom, self.reconPhantom)

        elif self.reconType in (ReconType.Algebraic, ReconType.Bayesian):
            self.MSE = []
            for theta in self.reconPhantom:
                self.MSE.append(mean_squared_error(self.experiment.OpticImage.phantom, theta))
  
    def calculateSSIM(self):
        """
        Calculate the Structural Similarity Index (SSIM) of the reconstruction.

        Returns:
            ssim: float or list of floats, Structural Similarity Index of the reconstruction
        """

        if self.reconPhantom is None or self.reconPhantom == []:
            raise ValueError("Reconstructed phantom is empty. Run reconstruction first.")
    
        if self.reconType in (ReconType.Analytic, ReconType.DeepLearning):
            data_range = self.reconPhantom.max() - self.reconPhantom.min()
            self.SSIM = ssim(self.experiment.OpticImage.phantom, self.reconPhantom, data_range=data_range)

        elif self.reconType in (ReconType.Algebraic, ReconType.Bayesian):
            self.SSIM = []
            for theta in self.reconPhantom:
                data_range = theta.max() - theta.min()
                ssim_value = ssim(self.experiment.OpticImage.phantom, theta, data_range=data_range)
                self.SSIM.append(ssim_value)
 
    @staticmethod
    def load_recon(hdr_path):
        """
        Lit un fichier Interfile (.hdr) et son fichier binaire (.img) pour reconstruire une image comme le fait Vinci.
        
        Paramètres :
        ------------
        - hdr_path : chemin complet du fichier .hdr
        
        Retour :
        --------
        - image : tableau NumPy contenant l'image
        - header : dictionnaire contenant les métadonnées du fichier .hdr
        """
        header = {}
        with open(hdr_path, 'r') as f:
            for line in f:
                if ':=' in line:
                    key, value = line.split(':=', 1)  # s'assurer qu'on ne coupe que la première occurrence de ':='
                    key = key.strip().lower().replace('!', '')  # Nettoyage des caractères
                    value = value.strip()
                    header[key] = value
        
        # 📘 Obtenez le nom du fichier de données associé (le .img)
        data_file = header.get('name of data file')
        if data_file is None:
            raise ValueError(f"Impossible de trouver le fichier de données associé au fichier header {hdr_path}")
        
        img_path = os.path.join(os.path.dirname(hdr_path), data_file)
        
        # 📘 Récupérer la taille de l'image à partir des métadonnées
        shape = [int(header[f'matrix size [{i}]']) for i in range(1, 4) if f'matrix size [{i}]' in header]
        if shape and shape[-1] == 1:  # Si la 3e dimension est 1, on la supprime
            shape = shape[:-1]  # On garde (192, 240) par exemple
        
        if not shape:
            raise ValueError("Impossible de déterminer la forme de l'image à partir des métadonnées.")
        
        # 📘 Déterminez le type de données à utiliser
        data_type = header.get('number format', 'short float').lower()
        dtype_map = {
            'short float': np.float32,
            'float': np.float32,
            'int16': np.int16,
            'int32': np.int32,
            'uint16': np.uint16,
            'uint8': np.uint8
        }
        dtype = dtype_map.get(data_type)
        if dtype is None:
            raise ValueError(f"Type de données non pris en charge : {data_type}")
        
        # 📘 Ordre des octets (endianness)
        byte_order = header.get('imagedata byte order', 'LITTLEENDIAN').lower()
        endianess = '<' if 'little' in byte_order else '>'
        
        # 📘 Vérifie la taille réelle du fichier .img
        img_size = os.path.getsize(img_path)
        expected_size = np.prod(shape) * np.dtype(dtype).itemsize
        
        if img_size != expected_size:
            raise ValueError(f"La taille du fichier img ({img_size} octets) ne correspond pas à la taille attendue ({expected_size} octets).")
        
        # 📘 Lire les données binaires et les reformater
        with open(img_path, 'rb') as f:
            data = np.fromfile(f, dtype=endianess + np.dtype(dtype).char)
        
        image =  data.reshape(shape[::-1]) 
        
        # 📘 Rescale l'image si nécessaire
        rescale_slope = float(header.get('data rescale slope', 1))
        rescale_offset = float(header.get('data rescale offset', 0))
        image = image * rescale_slope + rescale_offset
        
        return image.T



from AOT_biomaps.Config import config

import os
import numpy as np
from concurrent.futures import ThreadPoolExecutor
from scipy.interpolate import RegularGridInterpolator
from scipy.signal import hilbert as np_hilbert
if config.get_process()  == 'gpu':
    import cupy as cp
    from cupyx.scipy.signal import hilbert as cp_hilbert
    import pynvml

def reshape_field(field,factor):
    """
    Downsample the acoustic field using interpolation to reduce its size for faster processing.
    This method uses interpolation to estimate values on a coarser grid.
    """
    try:
        if field is None:
            raise ValueError("Acoustic field is not generated. Please generate the field first.")

        if len(factor) == 3:
            # Create new grid for 3D field
            x = np.arange(field.shape[0])
            y = np.arange(field.shape[1])
            z = np.arange(field.shape[2])

            # Create interpolating function
            interpolator = RegularGridInterpolator((x, y, z), field)

            # Create new coarser grid points
            x_new = np.linspace(0, field.shape[0] - 1, field.shape[0] // factor[0])
            y_new = np.linspace(0, field.shape[1] - 1, field.shape[1] // factor[1])
            z_new = np.linspace(0, field.shape[2] - 1, field.shape[2] // factor[2])

            # Create meshgrid for new points
            x_grid, y_grid, z_grid = np.meshgrid(x_new, y_new, z_new, indexing='ij')

            # Interpolate values
            points = np.stack((x_grid.flatten(), y_grid.flatten(), z_grid.flatten()), axis=-1)
            smoothed_field = interpolator(points).reshape(x_grid.shape)

            return smoothed_field

        elif len(factor) == 4:
            # Create new grid for 4D field
            x = np.arange(field.shape[0])
            y = np.arange(field.shape[1])
            z = np.arange(field.shape[2])
            w = np.arange(field.shape[3])

            # Create interpolating function
            interpolator = RegularGridInterpolator((x, y, z, w), field)

            # Create new coarser grid points
            x_new = np.linspace(0, field.shape[0] - 1, field.shape[0] // factor[0])
            y_new = np.linspace(0, field.shape[1] - 1, field.shape[1] // factor[1])
            z_new = np.linspace(0, field.shape[2] - 1, field.shape[2] // factor[2])
            w_new = np.linspace(0, field.shape[3] - 1, field.shape[3] // factor[3])

            # Create meshgrid for new points
            x_grid, y_grid, z_grid, w_grid = np.meshgrid(x_new, y_new, z_new, w_new, indexing='ij')

            # Interpolate values
            points = np.stack((x_grid.flatten(), y_grid.flatten(), z_grid.flatten(), w_grid.flatten()), axis=-1)
            smoothed_field = interpolator(points).reshape(x_grid.shape)

            return smoothed_field

        else:
            raise ValueError("Invalid dimension for downsampling. Supported dimensions are: 3D, 4D.")

    except Exception as e:
        print(f"Error in interpolate_reshape_field method: {e}")
        raise

def calculate_envelope_squared(field, isGPU= config.get_process() == 'gpu'):
    """
    Calculate the analytic envelope of the acoustic field using either CPU or GPU.

    Parameters:
    - use_gpu (bool): If True, use GPU for computation. Otherwise, use CPU.

    Returns:
    - envelope (numpy.ndarray or cupy.ndarray): The squared analytic envelope of the acoustic field.
    """
    try:                
        if field is None:
            raise ValueError("Acoustic field is not generated. Please generate the field first.")

        if isGPU:

            pynvml.nvmlInit()
            handle = pynvml.nvmlDeviceGetHandleByIndex(0)  # Assuming you want to check the first GPU
            info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            total_memory = info.total / (1024 ** 2)  # Convert to MB
            used_memory = info.used / (1024 ** 2)  # Convert to MB
            free_memory = int(total_memory - used_memory)
            
            if free_memory < field.nbytes / (1024 ** 2):
                print(f"GPU memory insufficient {int(field.nbytes / (1024 ** 2))} MB, Free GPU memory: {free_memory} MB, falling back to CPU.")
                isGPU = False
                acoustic_field = np.asarray(field)
            else:
                acoustic_field = cp.asarray(field)                   
        else:
            acoustic_field = np.asarray(field)

        if len(acoustic_field.shape) not in [3, 4]:
            raise ValueError("Input acoustic field must be a 3D or 4D array.")

        def process_slice(slice_index,isGPU):
            """Calculate the envelope for a given slice of the acoustic field."""
            if isGPU:
                hilbert = cp_hilbert
            else:
                hilbert = np_hilbert

            if len(acoustic_field.shape) == 3:
                return np.abs(hilbert(acoustic_field[slice_index], axis=0))**2
            elif len(acoustic_field.shape) == 4:
                envelope_slice = np.zeros_like(acoustic_field[slice_index])
                for y in range(acoustic_field.shape[2]):
                    for z in range(acoustic_field.shape[1]):
                        envelope_slice[:, z, y, :] = np.abs(hilbert(acoustic_field[slice_index][:, z, y, :], axis=0))**2
                return envelope_slice

        # Determine the number of slices to process in parallel
        num_slices = acoustic_field.shape[0]
        slice_indices = [(i,) for i in range(num_slices)]

        if isGPU:
            # Use GPU directly without multithreading
            envelopes = [process_slice(slice_index,isGPU) for slice_index in slice_indices]
        else:
            # Use ThreadPoolExecutor to parallelize the computation on CPU
            with ThreadPoolExecutor() as executor:
                envelopes = list(executor.map(lambda index: process_slice(index,isGPU), slice_indices))

        # Combine the results into a single array
        if isGPU:
            return cp.stack(envelopes, axis=0).get()
        else:
            return np.stack(envelopes, axis=0)

    except Exception as e:
        print(f"Error in calculate_envelope_squared method: {e}")
        raise

def getPattern(pathFile):
    """
    Get the pattern from a file path.

    Args:
        pathFile (str): Path to the file containing the pattern.

    Returns:
        str: The pattern string.
    """
    try:
        # Pattern between first _ and last _
        pattern = os.path.basename(pathFile).split('_')[1:-1]
        pattern_str = ''.join(pattern)
        return pattern_str
    except Exception as e:
        print(f"Error reading pattern from file: {e}")
        return None

def getAngle(pathFile):
    """
    Get the angle from a file path.

    Args:
        pathFile (str): Path to the file containing the angle.

    Returns:
        int: The angle in degrees.
    """
    try:
        # Angle between last _ and .
        angle_str = os.path.basename(pathFile).split('_')[-1].replace('.', '')
        if angle_str.startswith('0'):
            angle_str = angle_str[1:]
        elif angle_str.startswith('1'):
            angle_str = '-' + angle_str[1:]
        else:
            raise ValueError("Invalid angle format in file name.")
        return int(angle_str)
    except Exception as e:
        print(f"Error reading angle from file: {e}")
        return None

def next_power_of_2(n):
    """Calculate the next power of 2 greater than or equal to n."""
    return int(2 ** np.ceil(np.log2(n)))
        



import GPUtil
import psutil

class Config:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Config, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        if not hasattr(self, 'initialized'):
            gpus = GPUtil.getGPUs()
            if not gpus:
                self.process = 'cpu'
                self.bestGPU = None
            else:
                self.process = 'gpu'
                self.bestGPU = self.select_best_gpu()
                self.numGPUs = len(gpus)
                self.numCPUs = psutil.cpu_count(logical=False)
            self.availableMemory = 100-self.get_memory_usage()
            self.batchSize = self.calculate_batch_size()
            self.initialized = True

    def set_process(self, process):
        if process not in ['cpu', 'gpu']:
            raise ValueError("process must be 'cpu' or 'gpu'")
        self.process = process

    def get_process(self):
        return self.process
    
    def select_best_gpu(self):
        gpus = GPUtil.getGPUs()

        best_gpu = 0
        max_memory = 0

        for i, gpu in enumerate(gpus):
            # Obtenez la mémoire totale et utilisée pour chaque GPU
            total_memory = gpu.memoryTotal
            used_memory = gpu.memoryUsed
            available_memory = total_memory - used_memory

            # Sélectionnez le GPU avec le plus de mémoire disponible
            if available_memory > max_memory:
                max_memory = available_memory
                best_gpu = i
        return best_gpu
    

    def get_memory_usage(self):
        """Returns the current memory usage in percentage."""
        return psutil.virtual_memory().percent

    def calculate_batch_size(self,max_memory_usage=90, min_batch_size=1, max_batch_size=20):
        """Calculate dynamic batch size based on available memory."""

        # Adjust batch size based on available memory
        if self.availableMemory > max_memory_usage:
            return max_batch_size
        else:
            # Scale batch size linearly with available memory
            return max(min_batch_size, int((self.availableMemory / max_memory_usage) * max_batch_size))

config = Config()


from AOT_biomaps.AOT_Recon.ReconTools import _forward_projection, _backward_projection

import numba
import torch
import numpy as np
import os
from tqdm import trange

def _MLEM_GPU_basic(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
    """
    This method implements the MLEM algorithm using PyTorch for GPU acceleration.
    Parameters:
        SMatrix: 4D numpy array (time, z, x, nScans)
        y: 2D numpy array (time, nScans)
        numIterations: number of iterations for the MLEM algorithm
    """
    A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float32).cuda()  # shape: (T, Z, X, N)
    y_torch = torch.tensor(y, dtype=torch.float32).cuda()                # shape: (T, N)

    # Initialize variables
    T, Z, X, N = SMatrix.shape

    # flat
    A_flat = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, Z * X)     # shape: (T * N, Z * X)
    y_flat = y_torch.reshape(-1)                                          # shape: (T * N, )

    # Step 1: start from a strickly positive image theta^(0)
    theta_0 = torch.ones((Z, X), dtype=torch.float32, device='cuda')      # shape: (Z, X)
    matrix_theta_torch = [theta_0]
    # matrix_theta_from_gpu = []

    # Compute normalization factor: A^T * 1
    normalization_factor = A_matrix_torch.sum(dim=(0, 3))                # shape: (Z, X)
    normalization_factor_flat = normalization_factor.reshape(-1)         # shape: (Z * X, )

    if withTumor:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
    else:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
    # EM Algebraic update
    for _ in trange(numIterations, desc=description):

        theta_p = matrix_theta_torch[-1]                                 # shape: (Z, X)

        # Step 2: Forward projection of current estimate : q = A * theta + b (acc with GPU)
        theta_p_flat = theta_p.reshape(-1)                               # shape: (Z * X, )
        q_flat = A_flat @ theta_p_flat                                   # shape: (T * N, )

        # Step 3: Current error estimate : compute ratio e = m / q
        e_flat = y_flat / (q_flat + torch.finfo(torch.float32).tiny)                                # shape: (T * N, )

        # Step 4: Backward projection of the error estimate : c = A.T * e (acc with GPU)
        c_flat = A_flat.T @ e_flat                                       # shape: (Z * X, )

        # Step 5: Multiplicative update of current estimate
        theta_p_plus_1_flat = (theta_p_flat / (normalization_factor_flat + torch.finfo(torch.float32).tiny)) * c_flat

        matrix_theta_torch.append(theta_p_plus_1_flat.reshape(Z, X))      # shape: (Z, X)
    
    if not isSavingEachIteration:
        return matrix_theta_torch[-1]
    else:
        return [theta.cpu().numpy() for theta in matrix_theta_torch]
        
def _MLEM_CPU_basic(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
    """
    This method implements the MLEM algorithm using basic numpy operations.
    Parameters:
        SMatrix: 4D numpy array (time, z, x, nScans)
        y: 2D numpy array (time, nScans)
        numIterations: number of iterations for the MLEM algorithm
    """
    # Initialize variables
    q_p = np.zeros((SMatrix.shape[0], SMatrix.shape[3]))  # shape : (t, i)
    c_p = np.zeros((SMatrix.shape[1], SMatrix.shape[2]))  # shape : (z, x)

    # Step 1: start from a strickly positive image theta^(0)
    theta_p_0 = np.ones((SMatrix.shape[1], SMatrix.shape[2]))  # initial theta^(0)
    matrix_theta = [theta_p_0]  # store theta 

    # Compute normalization factor: A^T * 1
    normalization_factor = np.sum(SMatrix, axis=(0, 3))  # shape: (z, x)

    if withTumor:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITH TUMOR ---- processing on single CPU (basic) ----"
    else:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITHOUT TUMOR ---- processing on single CPU (basic) ----"
    
    # EM Algebraic update
    for _ in trange(numIterations, desc=description):

        theta_p = matrix_theta[-1]

        # Step 1: Forward projection of current estimate : q = A * theta + b
        for _t in range(SMatrix.shape[0]):
            for _n in range(SMatrix.shape[3]):
                q_p[_t, _n] = np.sum(SMatrix[_t, :, :, _n] * theta_p)
        
        # Step 2: Current error estimate : compute ratio e = m / q
        e_p = y / (q_p + 1e-8)  # 避免除零
        
        # Step 3: Backward projection of the error estimate : c = A.T * e
        for _z in range(SMatrix.shape[1]):
            for _x in range(SMatrix.shape[2]):
                c_p[_z, _x] = np.sum(SMatrix[:, _z, _x, :] * e_p)
        
        # Step 4: Multiplicative update of current estimate
        theta_p_plus_1 = theta_p / (normalization_factor + 1e-8) * c_p
        
        # Step 5: Store current theta
        matrix_theta.append(theta_p_plus_1)
    
    if not isSavingEachIteration:
        return matrix_theta[-1]
    else:
        return matrix_theta  # Return the list of numpy arrays

def _MLEM_CPU_multi(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
    """
    This method implements the MLEM algorithm using multi-threading with Numba.
    Parameters:
        SMatrix: 4D numpy array (time, z, x, nScans)
        y: 2D numpy array (time, nScans)
        numIterations: number of iterations for the MLEM algorithm
    """
    numba.set_num_threads(os.cpu_count())
    q_p = np.zeros((SMatrix.shape[0], SMatrix.shape[3]))  # shape : (t, i)
    c_p = np.zeros((SMatrix.shape[1], SMatrix.shape[2]))  # shape : (z, x)

    # Step 1: start from a strickly positive image theta^(0)
    theta_p_0 = np.ones((SMatrix.shape[1], SMatrix.shape[2]))
    matrix_theta = [theta_p_0]

    # Compute normalization factor: A^T * 1
    normalization_factor = np.sum(SMatrix, axis=(0, 3))  # shape: (z, x)

    if withTumor:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITH TUMOR ---- processing on multithread CPU ({numba.config.NUMBA_DEFAULT_NUM_THREADS} threads)----"
    else:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITHOUT TUMOR ---- processing on multithread CPU ({numba.config.NUMBA_DEFAULT_NUM_THREADS} threads)----"

    # EM Algebraic update
    for _ in trange(numIterations, desc=description):
        
        theta_p = matrix_theta[-1]

        # Step 1: Forward projection of current estimate : q = A * theta + b (acc with njit)
        _forward_projection(SMatrix, theta_p, q_p)

        # Step 2: Current error estimate : compute ratio e = m / q
        e_p = y / (q_p + 1e-8)

        # Step 3: Backward projection of the error estimate : c = A.T * e (acc with njit)
        _backward_projection(SMatrix, e_p, c_p)

        # Step 4: Multiplicative update of current estimate
        theta_p_plus_1 = theta_p / (normalization_factor + 1e-8) * c_p

        # Step 5: Store current theta
        matrix_theta.append(theta_p_plus_1)
    
    if not isSavingEachIteration:
        return matrix_theta[-1]
    else:
        return matrix_theta

def _MLEM_CPU_opti(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
    """
    This method implements the MLEM algorithm using optimized numpy operations.
    Parameters:
        SMatrix: 4D numpy array (time, z, x, nScans)
        y: 2D numpy array (time, nScans)
        numIterations: number of iterations for the MLEM algorithm
    """
    # Initialize variables
    T, Z, X, N = SMatrix.shape

    A_flat = SMatrix.astype(np.float32).transpose(0, 3, 1, 2).reshape(T * N, Z * X)         # shape: (T * N, Z * X)
    y_flat = y.astype(np.float32).reshape(-1)                                                # shape: (T * N, )

    # Step 1: start from a strickly positive image theta^(0)
    theta_0 = np.ones((Z, X), dtype=np.float32)              # shape: (Z, X)
    matrix_theta = [theta_0]

    # Compute normalization factor: A^T * 1
    normalization_factor = np.sum(SMatrix, axis=(0, 3)).astype(np.float32)                  # shape: (Z, X)
    normalization_factor_flat = normalization_factor.reshape(-1) 

    if withTumor:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITH TUMOR ---- processing on single CPU (optimized) ----"
    else:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITHOUT TUMOR ---- processing on single CPU (optimized) ----"

    # EM Algebraic update
    for _ in trange(numIterations, desc=description):
        theta_p = matrix_theta[-1]

        # Step 2: Forward projection of current estimate : q = A * theta + b (acc with njit)
        theta_p_flat = theta_p.reshape(-1)                                                    # shape: (Z * X, )
        q_flat = A_flat @ theta_p_flat                                                        # shape: (T * N)

        # Step 3: Current error estimate : compute ratio e = m / q
        e_flat = y_flat / (q_flat + np.finfo(np.float32).tiny)                                         # shape: (T * N, )
        # np.float32(1e-8)
        
        # Step 4: Backward projection of the error estimate : c = A.T * e (acc with njit)
        c_flat = A_flat.T @ e_flat                                                            # shape: (Z * X, )

        # Step 5: Multiplicative update of current estimate
        theta_p_plus_1_flat = theta_p_flat / (normalization_factor_flat + np.finfo(np.float32).tiny) * c_flat
        
        
        # Step 5: Store current theta
        matrix_theta.append(theta_p_plus_1_flat.reshape(Z, X))

    if not isSavingEachIteration:
        return matrix_theta[-1]
    else:
        return matrix_theta


def _MLEM_GPU_multi(SMatrix, y, numIterations, isSavingEachIteration, withTumor):
    """
    This method implements the MLEM algorithm using PyTorch for multi-GPU acceleration.
    Parameters:
        SMatrix: 4D numpy array (time, z, x, nScans)
        y: 2D numpy array (time, nScans)
        numIterations: number of iterations for the MLEM algorithm
    """
    # Check the number of available GPUs
    num_gpus = torch.cuda.device_count()

    # Convert data to tensors and distribute across GPUs
    A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float32).cuda()
    y_torch = torch.tensor(y, dtype=torch.float32).cuda()

    # Initialize variables
    T, Z, X, N = SMatrix.shape

    # Distribute the data across GPUs
    A_matrix_torch = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y_torch.reshape(-1)

    # Split data across GPUs
    A_split = torch.chunk(A_matrix_torch, num_gpus, dim=0)
    y_split = torch.chunk(y_flat, num_gpus)

    # Initialize theta on each GPU
    theta_0 = torch.ones((Z, X), dtype=torch.float32, device='cuda')
    theta_list = [theta_0.clone() for _ in range(num_gpus)]

    # Compute normalization factor: A^T * 1
    normalization_factor = A_matrix_torch.sum(dim=0)
    normalization_factor = normalization_factor.reshape(Z, X)

    if withTumor:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITH TUMOR ---- processing on multi-GPU ({num_gpus} GPUs)----"
    else:
        description = f"AOT-BioMaps -- Algebraic Reconstruction Tomography: ML-EM  ---- WITHOUT TUMOR ---- processing on multi-GPU ({num_gpus} GPUs)----"

    # EM Algebraic update
    for _ in trange(numIterations, desc=description):
        theta_p_list = theta_list.copy()

        for i in range(num_gpus):
            A_i = A_split[i].to(f'cuda:{i}')
            y_i = y_split[i].to(f'cuda:{i}')
            theta_p = theta_p_list[i].to(f'cuda:{i}')

            # Forward projection
            q_flat = A_i @ theta_p.reshape(-1)

            # Current error estimate
            e_flat = y_i / (q_flat + torch.finfo(torch.float32).tiny)

            # Backward projection
            c_flat = A_i.T @ e_flat

            # Multiplicative update
            theta_p_plus_1_flat = (theta_p.reshape(-1) / (normalization_factor.reshape(-1) + torch.finfo(torch.float32).tiny)) * c_flat
            theta_list[i] = theta_p_plus_1_flat.reshape(Z, X).to('cuda:0')

    if not isSavingEachIteration:
        return torch.stack(theta_list).mean(dim=0).cpu().numpy()
    else:
        return [theta.cpu().numpy() for theta in theta_list]



from AOT_biomaps.AOT_Recon.ReconEnums import PotentialType
from AOT_biomaps.AOT_Recon.AOT_PotentialFunctions.Quadratic import _Omega_QUADRATIC_CPU, _Omega_QUADRATIC_GPU
from AOT_biomaps.AOT_Recon.AOT_PotentialFunctions.RelativeDifferences import _Omega_RELATIVE_DIFFERENCE_CPU, _Omega_RELATIVE_DIFFERENCE_GPU
from AOT_biomaps.AOT_Recon.AOT_PotentialFunctions.Huber import _Omega_HUBER_PIECEWISE_CPU, _Omega_HUBER_PIECEWISE_GPU
from AOT_biomaps.AOT_Recon.ReconTools import _build_adjacency_sparse_CPU, _build_adjacency_sparse_GPU
from AOT_biomaps.Config import config

import numpy as np
import torch
from tqdm import trange


def _MAPEM_CPU_STOP(SMatrix, y, Omega, numIterations, beta, delta, gamma, sigma, isSavingEachIteration, withTumor):
    """
    MAPEM version CPU simple - sans GPU - torch uniquement
    """
    if Omega is not isinstance(Omega, PotentialType):
        raise TypeError(f"Omega must be of type PotentialType, got {type(Omega)}")

    if Omega == PotentialType.RELATIVE_DIFFERENCE:
        if gamma == None:
            raise ValueError("gamma must be specified for RELATIVE_DIFFERENCE potential type. Please find the value in the paper.")
        if beta == None:
            raise ValueError("beta must be specified for RELATIVE_DIFFERENCE potential type. Please find the value in the paper.")
    elif Omega == PotentialType.HUBER_PIECEWISE:
        if delta == None:
            raise ValueError("delta must be specified for HUBER_PIECEWISE potential type. Please find the value in the paper.")
        if beta == None:
            raise ValueError("beta must be specified for HUBER_PIECEWISE potential type. Please find the value in the paper.") 
    elif Omega == PotentialType.QUADRATIC:
        if sigma == None:
            raise ValueError("sigma must be specified for QUADRATIC potential type. Please find the value in the paper.")
    else:
        raise ValueError(f"Unknown potential type: {Omega}")
    
    SMatrix = torch.tensor(SMatrix, dtype=torch.float32)
    y_tensor = torch.tensor(y, dtype=torch.float32)

    T, Z, X, N = SMatrix.shape
    A_flat = SMatrix.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y_tensor.reshape(-1)

    I_0 = torch.ones((Z, X), dtype=torch.float32)
    theta_list = [I_0]
    results = [I_0.numpy()]
    previous=-np.inf

    normalization_factor = SMatrix.sum(dim=(0, 3)).reshape(-1)
    adj_index, adj_values = _build_adjacency_sparse_CPU(Z, X)

    if Omega == PotentialType.RELATIVE_DIFFERENCE:
            if withTumor:
                description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f})+ STOP condtion (penalized log-likelihood) ---- WITH TUMOR ---- processing on single CPU ----"
            else:
                description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f})+ STOP condtion (penalized log-likelihood) ---- WITHOUT TUMOR ---- processing on single CPU ----"
    elif Omega == PotentialType.HUBER_PIECEWISE:
            if withTumor:
                description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f})+ STOP condtion (penalized log-likelihood) ---- WITH TUMOR ---- processing on single CPU ----"
            else:
                description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f})+ STOP condtion (penalized log-likelihood) ---- WITHOUT TUMOR ---- processing on single CPU ----"
    elif Omega == PotentialType.QUADRATIC:
            if withTumor:
                description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f})+ STOP condtion (penalized log-likelihood) ---- WITH TUMOR ---- processing on single CPU ----"
            else:
                description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f})+ STOP condtion (penalized log-likelihood) ---- WITHOUT TUMOR ---- processing on single CPU ----"

    for p in trange(numIterations, desc=description):
        theta_p = theta_list[-1]
        theta_p_flat = theta_p.reshape(-1)

        q_flat = A_flat @ theta_p_flat
        e_flat = (y_flat - q_flat) / (q_flat + torch.finfo(torch.float32).tiny)
        c_flat = A_flat.T @ e_flat

        if Omega == PotentialType.RELATIVE_DIFFERENCE:
            grad_U, hess_U, U_value = _Omega_RELATIVE_DIFFERENCE_CPU(theta_p_flat, adj_index, adj_values, delta)
        elif Omega == PotentialType.HUBER_PIECEWISE:
            grad_U, hess_U, U_value = _Omega_HUBER_PIECEWISE_CPU(theta_p_flat, adj_index, adj_values, gamma)
        elif Omega == PotentialType.QUADRATIC:
            grad_U, hess_U, U_value = _Omega_QUADRATIC_CPU(theta_p_flat, adj_index, adj_values, sigma)

        denom = normalization_factor + theta_p_flat * beta * hess_U
        num = theta_p_flat * (c_flat - beta * grad_U)

        theta_next_flat = theta_p_flat + num / (denom + torch.finfo(torch.float32).tiny)
        theta_next_flat = torch.clamp(theta_next_flat, min=0)
        theta_next = theta_next_flat.reshape(Z, X)

        theta_list[-1] = theta_next
        results.append(theta_next.numpy())

        log_likelihood = (y_flat * torch.log(q_flat + 1e-8) - (q_flat + 1e-8)).sum()
        penalized_log_likelihood = log_likelihood - beta * U_value
        current = penalized_log_likelihood.item()

        if (p + 1) % 100 == 0:
            print(f"Iter {p+1}: logL={log_likelihood:.3e}, U={U_value:.3e}, penalized logL={penalized_log_likelihood:.3e}")
            if current <= previous:
                nb_false_successive += 1
            else:
                nb_false_successive = 0
            previous = current

            # Optionally add early stop:
            # if nb_false_successive >= 25:
            #     break

    if isSavingEachIteration:
        return results[-1]
    else:
        return results  

def _MAPEM_GPU_STOP(SMatrix, y, Omega, numIterations, beta, delta, gamma, sigma, isSavingEachIteration, withTumor):
    """
    Maximum A Posteriori (MAP) estimation for Bayesian reconstruction.
    This method computes the MAP estimate of the parameters given the data.
    """

    if Omega is not isinstance(Omega, PotentialType):
        raise TypeError(f"Omega must be of type PotentialType, got {type(Omega)}")

    if Omega == PotentialType.RELATIVE_DIFFERENCE:
        if gamma == None:
            raise ValueError("gamma must be specified for RELATIVE_DIFFERENCE potential type. Please find the value in the paper.")
        if beta == None:
            raise ValueError("beta must be specified for RELATIVE_DIFFERENCE potential type. Please find the value in the paper.")
    elif Omega == PotentialType.HUBER_PIECEWISE:
        if delta == None:
            raise ValueError("delta must be specified for HUBER_PIECEWISE potential type. Please find the value in the paper.")
        if beta == None:
            raise ValueError("beta must be specified for HUBER_PIECEWISE potential type. Please find the value in the paper.") 
    elif Omega == PotentialType.QUADRATIC:
        if sigma == None:
            raise ValueError("sigma must be specified for QUADRATIC potential type. Please find the value in the paper.")
    else:
        raise ValueError(f"Unknown potential type: {Omega}")

    device = torch.device(f"cuda:{config.select_best_gpu()}")

    A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float32).to(device)
    y_torch = torch.tensor(y, dtype=torch.float32).to(device)

    T, Z, X, N = SMatrix.shape
    J = Z * X

    A_flat = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y_torch.reshape(-1)

    I_0 = torch.ones((Z, X), dtype=torch.float32, device=device)
    matrix_theta_torch = []
    matrix_theta_torch = [I_0]
    matrix_theta_from_gpu_MAPEM = []
    matrix_theta_from_gpu_MAPEM = [I_0.cpu().numpy()]

    normalization_factor = A_matrix_torch.sum(dim=(0, 3))                # (Z, X)
    normalization_factor_flat = normalization_factor.reshape(-1)         # (Z*X,)
    previous = -np.inf
    adj_index, adj_values = _build_adjacency_sparse_GPU(Z, X)

    
    if Omega == PotentialType.RELATIVE_DIFFERENCE:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f})+ STOP condtion (penalized log-likelihood) ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f})+ STOP condtion (penalized log-likelihood) ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
    elif Omega == PotentialType.HUBER_PIECEWISE:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f})+ STOP condtion (penalized log-likelihood) ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f})+ STOP condtion (penalized log-likelihood) ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
    elif Omega == PotentialType.QUADRATIC:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f})+ STOP condtion (penalized log-likelihood) ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"   
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f})+ STOP condtion (penalized log-likelihood) ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"

    for p in trange(numIterations, desc=description):
        theta_p = matrix_theta_torch[-1]
        theta_p_flat = theta_p.reshape(-1)

        q_flat = A_flat @ theta_p_flat
        e_flat = (y_flat - q_flat) / (q_flat + torch.finfo(torch.float32).tiny)
        c_flat = A_flat.T @ e_flat

        if Omega == PotentialType.RELATIVE_DIFFERENCE:
                grad_U, hess_U, U_value = _Omega_RELATIVE_DIFFERENCE_CPU(theta_p_flat, adj_index, adj_values, delta)
        elif Omega == PotentialType.HUBER_PIECEWISE:
                grad_U, hess_U, U_value = _Omega_HUBER_PIECEWISE_CPU(theta_p_flat, adj_index, adj_values, gamma)
        elif Omega == PotentialType.QUADRATIC:
                grad_U, hess_U, U_value = _Omega_QUADRATIC_CPU(theta_p_flat, adj_index, adj_values, sigma)
        else:
            raise ValueError(f"Unknown potential type: {Omega}")
        
        denom = normalization_factor_flat + theta_p_flat * beta * hess_U
        num = theta_p_flat * (c_flat - beta * grad_U)

        theta_p_plus_1_flat = theta_p_flat + num / (denom + torch.finfo(torch.float32).tiny)
        theta_p_plus_1_flat = torch.clamp(theta_p_plus_1_flat, min=0)

        theta_next = theta_p_plus_1_flat.reshape(Z, X)
        #matrix_theta_torch.append(theta_next) # save theta in GPU
        matrix_theta_torch[-1] = theta_next    # do not save theta in GPU

        if p % 1 == 0:
            matrix_theta_from_gpu_MAPEM.append(theta_next.cpu().numpy())

        # === compute penalized log-likelihood (without term ln(m_i !) inside) ===
        # log-likelihood (without term ln(m_i !) inside)
        # log_likelihood = (torch.where(q_flat > 0, y_flat * torch.log(q_flat), torch.zeros_like(q_flat)) - q_flat).sum()
        # log_likelihood = (y_flat * torch.log(q_flat) - q_flat).sum()
        log_likelihood = (y_flat * ( torch.log( q_flat + torch.finfo(torch.float32).tiny ) ) - (q_flat + torch.finfo(torch.float32).tiny)).sum()

        # penalized log-likelihood
        penalized_log_likelihood = log_likelihood - beta * U_value

        if p == 0 or (p+1) % 100 == 0:
            current = penalized_log_likelihood.item()

            if current<=previous:
                nb_false_successive = nb_false_successive + 1

            else:
                nb_false_successive = 0
            
            print(f"Iter {p+1}: lnL without term ln(m_i !) inside={log_likelihood.item():.8e}, Gibbs energy function U={U_value.item():.4e}, penalized lnL without term ln(m_i !) inside={penalized_log_likelihood.item():.8e}, p lnL (current {current:.8e} - previous {previous:.8e} > 0)={(current-previous>0)}, nb_false_successive={nb_false_successive}")
            
            #if nb_false_successive >= 25:
                #break
        
            previous = penalized_log_likelihood.item()
    if isSavingEachIteration:
        return matrix_theta_from_gpu_MAPEM[-1]
    else:
        return matrix_theta_from_gpu_MAPEM

def _MAPEM_CPU(SMatrix, y, Omega, numIterations, beta, delta, gamma, sigma, isSavingEachIteration, withTumor):
    if not isinstance(Omega, PotentialType):
        raise TypeError(f"Omega must be of type PotentialType, got {type(Omega)}")

    if Omega == PotentialType.RELATIVE_DIFFERENCE:
        if gamma is None or beta is None:
            raise ValueError("gamma and beta must be specified for RELATIVE_DIFFERENCE.")
    elif Omega == PotentialType.HUBER_PIECEWISE:
        if delta is None or beta is None:
            raise ValueError("delta and beta must be specified for HUBER_PIECEWISE.")
    elif Omega == PotentialType.QUADRATIC:
        if sigma is None:
            raise ValueError("sigma must be specified for QUADRATIC.")
    else:
        raise ValueError(f"Unknown potential type: {Omega}")

    T, Z, X, N = SMatrix.shape
    J = Z * X

    A_flat = np.transpose(SMatrix, (0, 3, 1, 2)).reshape(T * N, Z * X)
    y_flat = y.reshape(-1)

    theta_0 = np.ones((Z, X), dtype=np.float32)
    matrix_theta_np = [theta_0]
    I_reconMatrix = [theta_0.copy()]

    normalization_factor = SMatrix.sum(axis=(0, 3))
    normalization_factor_flat = normalization_factor.reshape(-1)

    adj_index, adj_values = _build_adjacency_sparse_CPU(Z, X)

    if Omega == PotentialType.RELATIVE_DIFFERENCE:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f}) ---- WITH TUMOR ---- processing on single CPU ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f}) ---- WITHOUT TUMOR ---- processing on single CPU ----"
    elif Omega == PotentialType.HUBER_PIECEWISE:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f}) ---- WITH TUMOR ---- processing on single CPU ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f}) ---- WITHOUT TUMOR ---- processing on single CPU ----"
    elif Omega == PotentialType.QUADRATIC:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITH TUMOR ---- processing on single CPU ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITHOUT TUMOR ---- processing on single CPU ----"

    for p in trange(numIterations, desc=description):
        theta_p = matrix_theta_np[-1]
        theta_p_flat = theta_p.reshape(-1)

        q_flat = A_flat @ theta_p_flat
        e_flat = (y_flat - q_flat) / (q_flat + np.finfo(np.float32).tiny)
        c_flat = A_flat.T @ e_flat

        if Omega == PotentialType.RELATIVE_DIFFERENCE:
            grad_U, hess_U, _ = _Omega_RELATIVE_DIFFERENCE_CPU(theta_p_flat, adj_index, adj_values, delta)
        elif Omega == PotentialType.HUBER_PIECEWISE:
            grad_U, hess_U, _ = _Omega_HUBER_PIECEWISE_CPU(theta_p_flat, adj_index, adj_values, gamma)
        elif Omega == PotentialType.QUADRATIC:
            grad_U, hess_U, _ = _Omega_QUADRATIC_CPU(theta_p_flat, adj_index, adj_values, sigma)


        denom = normalization_factor_flat + theta_p_flat * beta * hess_U
        num = theta_p_flat * (c_flat - beta * grad_U)

        theta_p_plus_1_flat = theta_p_flat + num / (denom + np.finfo(np.float32).tiny)
        theta_p_plus_1_flat = np.clip(theta_p_plus_1_flat, 0, None)

        theta_next = theta_p_plus_1_flat.reshape(Z, X)
        matrix_theta_np.append(theta_next)


        if p % 1 == 0:
            I_reconMatrix.append(theta_next.copy())

    if isSavingEachIteration:
        return I_reconMatrix
    else:
        return I_reconMatrix[-1]

def _MAPEM_GPU(SMatrix, y, Omega, numIterations, beta, delta, gamma, sigma, isSavingEachIteration, withTumor):
    '''
    Maximum A Posteriori (MAP) estimation for Bayesian reconstruction using GPU.
    This method computes the MAP estimate of the parameters given the data.
    Parameters:
        SMatrix (numpy.ndarray): The system matrix of shape (T, Z, X, N).
        y (numpy.ndarray): The observed data of shape (T, N).
        Omega (PotentialType): The potential function to use for regularization.
        iteration (int): The number of iterations for the MAP-EM algorithm.
    Returns:
        matrix_theta_from_gpu_MAPEM (list): A list of numpy arrays containing the estimated parameters at each iteration.
        '''
    
    if Omega is not isinstance(Omega, PotentialType):
        raise TypeError(f"Omega must be of type PotentialType, got {type(Omega)}")

    if Omega == PotentialType.RELATIVE_DIFFERENCE:
        if gamma == None:
            raise ValueError("gamma must be specified for RELATIVE_DIFFERENCE potential type. Please find the value in the paper.")
        if beta == None:
            raise ValueError("beta must be specified for RELATIVE_DIFFERENCE potential type. Please find the value in the paper.")
    elif Omega == PotentialType.HUBER_PIECEWISE:
        if delta == None:
            raise ValueError("delta must be specified for HUBER_PIECEWISE potential type. Please find the value in the paper.")
        if beta == None:
            raise ValueError("beta must be specified for HUBER_PIECEWISE potential type. Please find the value in the paper.") 
    elif Omega == PotentialType.QUADRATIC:
        if sigma == None:
            raise ValueError("sigma must be specified for QUADRATIC potential type. Please find the value in the paper.")
    else:
        raise ValueError(f"Unknown potential type: {Omega}")

    device = torch.device(f"cuda:{config.select_best_gpu()}")

    A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float32).to(device)
    y_torch = torch.tensor(y, dtype=torch.float32).to(device)

    T, Z, X, N = SMatrix.shape
    J = Z * X

    A_flat = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y_torch.reshape(-1)

    theta_0 = torch.ones((Z, X), dtype=torch.float32, device=device)
    matrix_theta_torch = []
    matrix_theta_torch = [theta_0]
    I_reconMatrix = [theta_0.cpu().numpy()]

    normalization_factor = A_matrix_torch.sum(dim=(0, 3))                # (Z, X)
    normalization_factor_flat = normalization_factor.reshape(-1)         # (Z*X,)

    adj_index, adj_values = _build_adjacency_sparse_GPU(Z, X)

    if Omega == PotentialType.RELATIVE_DIFFERENCE:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f}) ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse RD β:{beta:.4f}, δ:{delta:4f}) ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"   
    elif Omega == PotentialType.HUBER_PIECEWISE:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f}) ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse HUBER β:{beta:.4f}, γ:{gamma:4f}) ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
    elif Omega == PotentialType.QUADRATIC:
        if withTumor:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
        else:
            description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : MAP-EM (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"

    for p in trange(numIterations, desc=description):
        theta_p = matrix_theta_torch[-1]
        theta_p_flat = theta_p.reshape(-1)

        q_flat = A_flat @ theta_p_flat
        e_flat = (y_flat - q_flat) / (q_flat + torch.finfo(torch.float32).tiny)
        c_flat = A_flat.T @ e_flat
        
        if Omega == PotentialType.RELATIVE_DIFFERENCE:
                grad_U, hess_U, _ = _Omega_RELATIVE_DIFFERENCE_GPU(theta_p_flat, adj_index, adj_values, delta=delta)
        elif Omega == PotentialType.HUBER_PIECEWISE:
                grad_U, hess_U, _ = _Omega_HUBER_PIECEWISE_GPU(theta_p_flat, adj_index, adj_values, gamma=gamma)
        elif Omega == PotentialType.QUADRATIC:
                grad_U, hess_U, _ = _Omega_QUADRATIC_GPU(theta_p_flat, adj_index, adj_values, sigma=sigma)
        else:
            raise ValueError(f"Unknown potential type: {Omega}")

        denom = normalization_factor_flat + theta_p_flat * beta * hess_U
        num = theta_p_flat * (c_flat - beta * grad_U)

        theta_p_plus_1_flat = theta_p_flat + num / (denom + torch.finfo(torch.float32).tiny)
        theta_p_plus_1_flat = torch.clamp(theta_p_plus_1_flat, min=0)

        theta_next = theta_p_plus_1_flat.reshape(Z, X)
        matrix_theta_torch.append(theta_next) # save theta in GPU

        if p % 1 == 0:
            I_reconMatrix.append(theta_next.cpu().numpy())   

    if isSavingEachIteration:
        return I_reconMatrix
    else:
        return I_reconMatrix[-1]



from AOT_biomaps.AOT_Acoustic._mainAcoustic import AcousticField
from .AcousticEnums import WaveType, TypeSim

import numpy as np

class IrregularWave(AcousticField):
    """
    Class for irregular wave types, inheriting from AcousticField.
    This class is a placeholder for future implementation of irregular wave types.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.waveType = WaveType.IrregularWave
        self.params = {
            'typeSim': TypeSim.IRREGULAR.value,
        }

    def getName_field(self):
        raise NotImplementedError("getName_field method not implemented for IrregularWave.")
    
    def  _generate_diverse_structurations(self,num_elements, num_sequences, num_frequencies):
        """
        Génère num_sequences structurations irrégulières ON/OFF pour une sonde de num_elements éléments.
        Chaque structuration contient exactement num_frequencies fréquences spatiales distinctes.

        :param num_elements: Nombre total d'éléments piézoélectriques de la sonde.
        :param num_sequences: Nombre total de structurations générées.
        :param num_frequencies: Nombre de fréquences spatiales distinctes par structuration.
        :return: Matrice de structuration de taille (num_sequences, num_elements)
        """
        
        # Définition des fréquences spatiales disponibles
        max_freq = num_elements // 2  # Nyquist limit
        available_frequencies = np.arange(1, max_freq + 1)  # Fréquences possibles
        
        # Matrice des structurations
        structurations = np.zeros((num_sequences, num_elements), dtype=int)
        
        # Sélectionner des fréquences uniques pour chaque structuration
        chosen_frequencies = []
        for _ in range(num_sequences):
            freqs = np.random.choice(available_frequencies, size=num_frequencies, replace=False)
            chosen_frequencies.append(freqs)

            # Construire la structuration correspondante
            structuration = np.zeros(num_elements)
            for f in freqs:
                structuration += np.cos(2 * np.pi * f * np.arange(num_elements) / num_elements)  # Ajouter la fréquence
            
            structuration = np.where(structuration >= 0, 1, 0)  # Binarisation ON/OFF
            structurations[_] = structuration
        
        return structurations, chosen_frequencies
    
    def getName_field(self):
        raise NotImplementedError("getName_field method not implemented for IrregularWave.")

    def _generate_2Dacoustic_field_KWAVE(self):
        raise NotImplementedError("2D acoustic field generation not implemented for IrregularWave.")

    def _generate_3Dacoustic_field_KWAVE(self):
        raise NotImplementedError("3D acoustic field generation not implemented for IrregularWave.")

    def _save2D_HDR_IMG(self, filePath):
        raise NotImplementedError("HDR/IMG saving not implemented for IrregularWave.")



from setuptools import setup, find_packages

setup(
    name='AOT_biomaps',
    version='2.7.9',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'k-wave-python',
        'setuptools',
        'pyyaml',
        'numba',
        'tqdm',
        'GPUtil',
        'scikit-image',
        'scikit-learn',
        'pandas',
    ],
    extras_require={
        'cpu': [
        ],
        'gpu': [
            'cupy',
            'nvidia-ml-py3',
            'torch',
        ],
    },
    author='Lucas Duclos',
    author_email='lucas.duclos@universite-paris-saclay.fr',
    description='Acousto-Optic Tomography',
    url='https://github.com/LucasDuclos/AcoustoOpticTomography',
)


from AOT_biomaps.AOT_Recon.AlgebraicRecon import AlgebraicRecon
from AOT_biomaps.AOT_Recon.ReconEnums import ReconType, OptimizerType, PotentialType, ProcessType
from AOT_biomaps.Config import config
from .AOT_Optimizers import MAPEM, DEPIERRO

import warnings
import numpy as np


if config.get_process()  == 'gpu':
    try:
        from torch_scatter import scatter
    except ImportError:
        raise ImportError("torch_scatter and torch_sparse are required for GPU processing. Please install them using 'pip install torch-scatter torch-sparse' with correct link (follow instructions https://github.com/LucasDuclos/AcoustoOpticTomography/edit/main/README.md).")

class BayesianRecon(AlgebraicRecon):
    """
    This class implements the Bayesian reconstruction process.
    It currently does not perform any operations but serves as a template for future implementations.
    """
    def __init__(self, 
                opti = OptimizerType.PGC,
                potentialFunction = PotentialType.HUBER_PIECEWISE,  
                beta=None, 
                delta=None, 
                gamma=None, 
                sigma=None,
                corner = (0.5-np.sqrt(2)/4)/np.sqrt(2),
                face = 0.5-np.sqrt(2)/4, 
                **kwargs):
        super().__init__(**kwargs)
        self.reconType = ReconType.Bayesian
        self.potentialFunction = potentialFunction
        self.optimizer = opti
        self.beta = beta           
        self.delta = delta          # typical value is 0.1
        self.gamma = gamma          # typical value is 0.01
        self.sigma = sigma          # typical value is 1.0
        self.corner = corner        # typical value is (0.5-np.sqrt(2)/4)/np.sqrt(2)
        self.face = face            # typical value is 0.5-np.sqrt(2)/4 

        if not isinstance(self.potentialFunction, PotentialType):
            raise TypeError(f"Potential functions must be of type PotentialType, got {type(self.potentialFunction)}")  

    def run(self, processType=ProcessType.PYTHON, withTumor=True):
        """
        This method is a placeholder for the Bayesian reconstruction process.
        It currently does not perform any operations but serves as a template for future implementations.
        """
        if(processType == ProcessType.CASToR):
            self._bayesianReconCASToR(withTumor)
        elif(processType == ProcessType.PYTHON):
            self._bayesianReconPython(withTumor)
        else:
            raise ValueError(f"Unknown Bayesian reconstruction type: {processType}")
        
    def _bayesianReconCASToR(self, withTumor):
        raise NotImplementedError("CASToR Bayesian reconstruction is not implemented yet.")

    def _bayesianReconPython(self, withTumor):

        if withTumor:
            if self.experiment.AOsignal_withTumor is None:
                raise ValueError("AO signal with tumor is not available. Please generate AO signal with tumor the experiment first in the experiment object.")
            if self.optimizer.value ==  OptimizerType.PPGMLEM.value:
                self.reconPhantom = self._MAPEM_STOP(SMatrix=self.SMatrix, y=self.experiment.AOsignal_withTumor, withTumor=withTumor)
            elif self.optimizer.value == OptimizerType.PGC.value:
                self.reconPhantom = self._MAPEM(SMatrix=self.SMatrix, y=self.experiment.AOsignal_withTumor, withTumor=withTumor)
            elif self.optimizer.value == OptimizerType.DEPIERRO95.value:
                self.reconPhantom = self._DEPIERRO(SMatrix=self.SMatrix, y=self.experiment.AOsignal_withTumor, withTumor=withTumor)
            else:
                raise ValueError(f"Unknown optimizer type: {self.optimizer.value}")
        else:
            if self.experiment.AOsignal_withoutTumor is None:
                raise ValueError("AO signal without tumor is not available. Please generate AO signal without tumor the experiment first in the experiment object.")
            if self.optimizer.value ==  OptimizerType.PPGMLEM.value:
                self.reconLaser = self._MAPEM_STOP(SMatrix=self.SMatrix, y=self.experiment.AOsignal_withoutTumor, withTumor=withTumor)
            elif self.optimizer.value == OptimizerType.PGC.value:
                self.reconLaser = self._MAPEM(SMatrix=self.SMatrix, y=self.experiment.AOsignal_withoutTumor, withTumor=withTumor)
            elif self.optimizer.value == OptimizerType.DEPIERRO95.value:
                self.reconLaser = self._DEPIERRO(SMatrix=self.SMatrix, y=self.experiment.AOsignal_withoutTumor, withTumor=withTumor)
            else:
                raise ValueError(f"Unknown optimizer type: {self.optimizer.value}")

    def _MAPEM_STOP(self, SMatrix, y, withTumor):

        if self.isGPU and self.isMultiGPU:
            warnings.warn("Multi-GPU MAPEM_STOP is not implemented yet. Falling back to single GPU.")
            try:
                return MAPEM._MAPEM_GPU_STOP(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
            except:
                warnings.warn("Falling back to CPU implementation due to an error in GPU implementation.")
                return MAPEM._MAPEM_CPU_STOP(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if self.isGPU and not self.isMultiGPU:
            try:
                return MAPEM._MAPEM_GPU_STOP(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
            except:
                warnings.warn("Falling back to CPU implementation due to an error in GPU implementation.")
                return MAPEM._MAPEM_CPU_STOP(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if not self.isGPU and self.isMultiCPU:
            warnings.warn("Multi-CPU MAPEM_STOP is not implemented yet. Falling back to single CPU.")
            return MAPEM._MAPEM_CPU_STOP(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if not self.isGPU and not self.isMultiCPU:
            return MAPEM._MAPEM_CPU_STOP(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        
    def _MAPEM(self, SMatrix, y, withTumor):
        
        if self.isGPU and self.isMultiGPU:
            warnings.warn("Multi-GPU MAPEM_STOP is not implemented yet. Falling back to single GPU.")
            try:
                return MAPEM._MAPEM_GPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations ,beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
            except:
                warnings.warn("Falling back to CPU implementation due to an error in GPU implementation.")
                return MAPEM._MAPEM_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if self.isGPU and not self.isMultiGPU:
            try:
                return MAPEM._MAPEM_GPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
            except:
                warnings.warn("Falling back to CPU implementation due to an error in GPU implementation.")
                return MAPEM._MAPEM_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if not self.isGPU and self.isMultiCPU:
            warnings.warn("Multi-CPU MAPEM_STOP is not implemented yet. Falling back to single CPU.")
            return MAPEM._MAPEM_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if not self.isGPU and not self.isMultiCPU:
            return MAPEM._MAPEM_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, delta=self.delta, gamma=self.gamma, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)

    def _DEPIERRO(self, SMatrix, y, withTumor):
        if self.isGPU and self.isMultiGPU:
            warnings.warn("Multi-GPU MAPEM_STOP is not implemented yet. Falling back to single GPU.")
            try:
                return DEPIERRO._DEPIERRO_GPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
            except:
                warnings.warn("Falling back to CPU implementation due to an error in GPU implementation.")
                return DEPIERRO._DEPIERRO_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if self.isGPU and not self.isMultiGPU:
            try:
                return DEPIERRO._DEPIERRO_GPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
            except:
                warnings.warn("Falling back to CPU implementation due to an error in GPU implementation.")
                return DEPIERRO._DEPIERRO_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if not self.isGPU and self.isMultiCPU:
            warnings.warn("Multi-CPU MAPEM_STOP is not implemented yet. Falling back to single CPU.")
            return DEPIERRO._DEPIERRO_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)
        if not self.isGPU and not self.isMultiCPU:
            return DEPIERRO._DEPIERRO_CPU(SMatrix= SMatrix, y = y, Omega=self.potentialFunction, numIterations=self.numIterations, beta=self.beta, sigma=self.sigma, isSavingEachIteration=self.isSavingEachIteration, withTumor=withTumor)






from AOT_biomaps.Config import config

import numpy as np
import torch
from numba import njit
if config.get_process()  == 'gpu':
    import torch
    try:
        from torch_scatter import scatter
    except ImportError:
        raise ImportError("torch_scatter and torch_sparse are required for GPU processing. Please install them using 'pip install torch-scatter torch-sparse' with correct link (follow instructions https://github.com/LucasDuclos/AcoustoOpticTomography/edit/main/README.md).")


@njit
def _Omega_HUBER_PIECEWISE_CPU(theta_flat, index, values, delta):
    """
    Compute the gradient and Hessian of the Huber penalty function for sparse data.
    Parameters:
        theta_flat (torch.Tensor): Flattened parameter vector.
        index (torch.Tensor): Indices of the sparse matrix in COO format.
        values (torch.Tensor): Values of the sparse matrix in COO format.
        delta (float): Threshold for the Huber penalty.
    Returns:
        grad_U (torch.Tensor): Gradient of the penalty function.
        hess_U (torch.Tensor): Hessian of the penalty function.
        U_value (float): Value of the penalty function.
    """
    j_idx, k_idx = index
    diff = theta_flat[j_idx] - theta_flat[k_idx]
    abs_diff = np.abs(diff)

    # Huber penalty (potential function)
    psi_pair = np.where(abs_diff > delta,
                        delta * abs_diff - 0.5 * delta ** 2,
                        0.5 * diff ** 2)
    psi_pair = values * psi_pair

    # Huber gradient
    grad_pair = np.where(abs_diff > delta,
                            delta * np.sign(diff),
                            diff)
    grad_pair = values * grad_pair

    # Huber Hessian
    hess_pair = np.where(abs_diff > delta,
                            np.zeros_like(diff),
                            np.ones_like(diff))
    hess_pair = values * hess_pair

    grad_U = np.zeros_like(theta_flat)
    hess_U = np.zeros_like(theta_flat)

    np.add.at(grad_U, j_idx, grad_pair)
    np.add.at(hess_U, j_idx, hess_pair)

    # Total penalty energy
    U_value = 0.5 * np.sum(psi_pair)

    return grad_U, hess_U, U_value

def _Omega_HUBER_PIECEWISE_GPU(theta_flat, index, values, delta):
    """
    Compute the gradient and Hessian of the Huber penalty function for sparse data.
    Parameters:
        theta_flat (torch.Tensor): Flattened parameter vector.
        index (torch.Tensor): Indices of the sparse matrix in COO format.
        values (torch.Tensor): Values of the sparse matrix in COO format.
        delta (float): Threshold for the Huber penalty.
    Returns:
        grad_U (torch.Tensor): Gradient of the penalty function.
        hess_U (torch.Tensor): Hessian of the penalty function.
        U_value (float): Value of the penalty function.
    """
    
    j_idx, k_idx = index
    diff = theta_flat[j_idx] - theta_flat[k_idx]
    abs_diff = torch.abs(diff)

    # Huber penalty (potential function) 
    psi_pair = torch.where(abs_diff > delta,
                        delta * abs_diff - 0.5 * delta ** 2,
                        0.5 * diff ** 2)
    psi_pair = values * psi_pair  

    # Huber gradient
    grad_pair = torch.where(abs_diff > delta,
                            delta * torch.sign(diff),
                            diff)
    grad_pair = values * grad_pair

    # Huber Hessian
    hess_pair = torch.where(abs_diff > delta,
                            torch.zeros_like(diff),
                            torch.ones_like(diff))
    hess_pair = values * hess_pair

    grad_U = scatter(grad_pair, j_idx, dim=0, dim_size=theta_flat.shape[0], reduce='sum')
    hess_U = scatter(hess_pair, j_idx, dim=0, dim_size=theta_flat.shape[0], reduce='sum')

    # Total penalty energy
    U_value = 0.5 * psi_pair.sum()  

    return grad_U, hess_U, U_value



from AOT_biomaps.AOT_Recon.ReconEnums import PotentialType
from AOT_biomaps.AOT_Recon.ReconTools import _build_adjacency_sparse_CPU, _build_adjacency_sparse_GPU
from AOT_biomaps.Config import config

import numpy as np
import torch
from tqdm import trange
if config.get_process()  == 'gpu':
    try:
        from torch_scatter import scatter
    except ImportError:
        raise ImportError("torch_scatter and torch_sparse are required for GPU processing. Please install them using 'pip install torch-scatter torch-sparse' with correct link (follow instructions https://github.com/LucasDuclos/AcoustoOpticTomography/edit/main/README.md).")


def _DEPIERRO_GPU(SMatrix, y, Omega, numIterations, beta, sigma, isSavingEachIteration, withTumor):

    if Omega != PotentialType.QUADRATIC:
        raise ValueError("Depierro95 optimizer only supports QUADRATIC potential function.")
    if beta is None or sigma is None:
        raise ValueError("Depierro95 optimizer requires beta and sigma parameters.")

    device = torch.device(f"cuda:{config.select_best_gpu()}")

    A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float32).to(device)
    y_torch = torch.tensor(y, dtype=torch.float32).to(device)

    T, Z, X, N = SMatrix.shape
    J = Z * X

    A_flat = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y_torch.reshape(-1)

    theta_0 = torch.ones((Z, X), dtype=torch.float32, device=device)
    matrix_theta_torch = []
    matrix_theta_torch = [theta_0]
    I_reconMatrix = [theta_0.cpu().numpy()]

    normalization_factor = A_matrix_torch.sum(dim=(0, 3))                # (Z, X)
    normalization_factor_flat = normalization_factor.reshape(-1)         # (Z*X,)

    adj_index, adj_values = _build_adjacency_sparse_GPU(Z, X)

    if withTumor:
        description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : DE PIERRO (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITH TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"
    else:
        description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : DE PIERRO (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITHOUT TUMOR ---- processing on single GPU no.{torch.cuda.current_device()} ----"

    for p in trange(numIterations, desc = description):

        theta_p = matrix_theta_torch[-1]

        # Step 2: Forward projection of current estimate : q = A * theta + b (acc with GPU)
        theta_p_flat = theta_p.reshape(-1)                               # shape: (Z * X, )
        q_flat = A_flat @ theta_p_flat                                   # shape: (T * N, )

        # Step 3: Current error estimate : compute ratio e = m / q
        e_flat = y_flat / (q_flat + torch.finfo(torch.float32).tiny)     # shape: (T * N, )

        # Step 4: Backward projection of the error estimate : c = A.T * e (acc with GPU)
        c_flat = A_flat.T @ e_flat                                       # shape: (Z * X, )

        # Step 5: Multiplicative update of current estimate
        #theta_p_plus_1_flat = (theta_p_flat / (normalization_factor_flat)) * c_flat
        #theta_EM_p_flat = (theta_p_flat / (normalization_factor_flat)) * c_flat
        theta_EM_p_flat = (theta_p_flat) * c_flat

        # --- Compute alpha_j ---
        alpha_j = normalization_factor_flat                                                                     # (Z*X,)

        # --- Compute W_j = (1/σ²) * ∑ ω_{kj} ---
        W_j = scatter(adj_values, adj_index[0], dim=0, dim_size=J, reduce='sum') * (1.0 / sigma**2)             # (Z*X,)

        # --- Compute γ_j = θ_j W_j + ∑_{k in N_j} θ_k ω_{kj} ---
        theta_k = theta_p_flat[adj_index[1]]
        weighted_theta_k = theta_k * adj_values
        gamma_j = theta_p_flat * W_j + scatter(weighted_theta_k, adj_index[0], dim=0, dim_size=J, reduce='sum')  # (Z*X,)

        # --- Pierro update ---
        A = 2 * beta * W_j
        B = - beta * gamma_j + alpha_j
        C = - theta_EM_p_flat
        '''
        if (p+1)%100 ==0 :
            print(f"torch.max(torch.abs(beta * gamma_j + alpha_j)- normalization_factor_flat ) = {torch.max(torch.abs(beta * gamma_j + alpha_j )- normalization_factor_flat) }")
            #print(f"torch.max(torch.abs(-B + torch.abs(B))) = {torch.max(torch.abs(-B + torch.abs(B)))}")
            val1 = torch.sqrt(B ** 2 - 4 * A * C)
            val2 = torch.abs(B) * (1 + 0.5 * (- 4 * A * C / (B**2)))
            print(torch.max(torch.abs(val1 - val2)))
        '''
        theta_p_plus_1_flat = (- B + torch.sqrt(B ** 2 - 4 * A * C)) / (2 * A + torch.finfo(torch.float32).tiny) 
        theta_p_plus_1_flat = torch.clamp(theta_p_plus_1_flat, min=0)
        
        theta_next = theta_p_plus_1_flat.reshape(Z, X)
        if (p+1)%100 ==0 :
            print(torch.sum(torch.abs(theta_p_plus_1_flat - theta_EM_p_flat/normalization_factor_flat)))
        #matrix_theta_torch.append(theta_next) # save theta in GPU
        matrix_theta_torch[-1] = theta_next    # do not save theta in GPU

        if p % 1 == 0:
            I_reconMatrix.append(theta_next.cpu().numpy()) 
        
    if isSavingEachIteration:
        return I_reconMatrix
    else:
        return I_reconMatrix[-1]

def _DEPIERRO_CPU(SMatrix, y, Omega, numIterations, beta, sigma, isSavingEachIteration, withTumor):
    if Omega != PotentialType.QUADRATIC:
        raise ValueError("Depierro95 optimizer only supports QUADRATIC potential function.")
    if beta is None or sigma is None:
        raise ValueError("Depierro95 optimizer requires beta and sigma parameters.")

    A_matrix = np.array(SMatrix, dtype=np.float32)
    y_array = np.array(y, dtype=np.float32)
    T, Z, X, N = SMatrix.shape
    J = Z * X

    A_flat = A_matrix.transpose(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y_array.reshape(-1)
    theta_0 = np.ones((Z, X), dtype=np.float32)
    matrix_theta = [theta_0]
    I_reconMatrix = [theta_0.copy()]

    normalization_factor = A_matrix.sum(axis=(0, 3))
    normalization_factor_flat = normalization_factor.reshape(-1)

    adj_index, adj_values = _build_adjacency_sparse_CPU(Z, X)

    if withTumor:
        description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : DE PIERRO (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITH TUMOR ---- processing on single CPU ----"
    else:
        description = f"AOT-BioMaps -- Bayesian Recontruction Tomography : DE PIERRO (Sparse QUADRATIC σ:{sigma:.4f}) ---- WITHOUT TUMOR ---- processing on single CPU ----"

    for p in trange(numIterations, desc = description):
        theta_p = matrix_theta[-1]
        theta_p_flat = theta_p.reshape(-1)

        # Forward projection of current estimate: q = A * theta
        q_flat = np.dot(A_flat, theta_p_flat)

        # Current error estimate: compute ratio e = m / q
        e_flat = y_flat / (q_flat + np.finfo(np.float32).tiny)

        # Backward projection of the error estimate: c = A.T * e
        c_flat = np.dot(A_flat.T, e_flat)

        # Multiplicative update of current estimate
        theta_EM_p_flat = theta_p_flat * c_flat

        # Compute alpha_j
        alpha_j = normalization_factor_flat

        # Compute W_j = (1/σ²) * ∑ ω_{kj}
        W_j = np.bincount(adj_index[0], weights=adj_values, minlength=J) * (1.0 / sigma**2)

        # Compute γ_j = θ_j W_j + ∑_{k in N_j} θ_k ω_{kj}
        theta_k = theta_p_flat[adj_index[1]]
        weighted_theta_k = theta_k * adj_values
        gamma_j = theta_p_flat * W_j + np.bincount(adj_index[0], weights=weighted_theta_k, minlength=J)

        # Pierro update
        A = 2 * beta * W_j
        B = -beta * gamma_j + alpha_j
        C = -theta_EM_p_flat

        theta_p_plus_1_flat = (-B + np.sqrt(B**2 - 4 * A * C)) / (2 * A + np.finfo(np.float32).tiny)
        theta_p_plus_1_flat = np.clip(theta_p_plus_1_flat, a_min=0, a_max=None)

        theta_next = theta_p_plus_1_flat.reshape(Z, X)
        matrix_theta[-1] = theta_next

        if p % 1 == 0:
            I_reconMatrix.append(theta_next.copy())

    if isSavingEachIteration:
        return I_reconMatrix
    else:
        return I_reconMatrix[-1]
